<?php

/**
 * @file
 * Entity Lister class file.
 */

class EntityLister {

  public $config;
  public $delta;
  public $flag;
  public $hostItem;
  protected $total;
  protected $bundleFields;

  /**
   * Construct the list object.
   *
   * @param array $config
   *   The config options for the list, in an associative array.  See below.
   * @param int $delta
   *   The delta.
   * @param string $flag
   *   An identifier.  When this class is used for a field formatter, this will
   *   be a field_name.  In cases where the field is configured to support
   *   multiple values, all values in the field should have the same string set
   *   for the $flag, but each should have a different $delta.
   * @param array $host_item
   *   When this class is used for a field formatter, this array describes the
   *   item to which the field is attached. See below.
   *
   *   $config['type']                string
   *     Entity type (e.g. 'comment', etc).
   *   $config['bundle']              array or json encoded string
   *   $config['view_mode']           string
   *   $config['numitems']            '1', '2', etc
   *   $config['sort']                array
   *     An array defining one or more sort configurations. Each
   *     configuration is itself an array with keys for 'method' and
   *     'args'.  The 'method' for each configuration should be set to either
   *     'propertyOrderBy' or 'fieldOrderBy'.  For the args, see the docs for
   *     EntityFieldQuery::propertyOrderBy and EntityFieldQuery::fieldOrderBy
   *     at https://api.drupal.org.  See also the demo page and
   *     EntityLister::defaults() for a 'propertyOrderBy' example.
   *   $config['pager']               '0', '1', or '2'
   *     Set to '0' for none, '1' for standard pager, or '2' for AJAX pager.
   *   $config['pager_position']      '0', '1', or '2'
   *     Set to '0' for above, '1' for below, or '2' for both.
   *   $config['pager_st_offset']     string
   *     Scrolltop offset from container top.
   *   $config['pager_st_time']       string
   *     Scrolltop time.
   *   $config['cache']               '0' or '1'
   *     Set to '1' to enable caching.
   *   $config['get_total']           '0' or '1'
   *     Set to '1' to execute a separate query for the total count of items.
   *   $config['status']              'all', '0' or '1'
   *     Set status to 'all' to list all items, unpublished and published.
   *     Set status to '0' to list only unpublished items.
   *     Set status to '1' to list only published items.
   *   $config['template']            string
   *     Specify an alternate template.
   *   $config['item_list']           '0' or '1'
   *     Set to '1' to pass the items to theme('item_list').
   *   $config['tag']                 string
   *     Specify a tag for addTag().
   *   $config['property_condition']  array
   *     Define one or more property conditions.
   *   $config['field_condition']     array
   *     Define one or more field conditions.
   *
   * The $host_item array:
   *   $host_item[0] The host item's entity_id.
   *   $host_item[1] The host item's revision_id.
   *   $host_item[2] The host item's bundle.
   *   $host_item[3] The host item's entity type.
   */
  public function __construct($config = array(), $delta = 0, $flag = '', $host_item = array()) {

    $this->config = $config;
    $this->delta = $delta;
    $this->flag = $flag;
    $this->hostItem = $host_item;

  }

  /**
   * Apply default values to the config where needed.
   */
  protected function defaults() {

    $c = $this->config;

    $config['type'] = isset($c['type']) ? $c['type'] : 'node';
    $config['bundle'] = isset($c['bundle']) ? $c['bundle'] : array('article');
    $config['view_mode'] = isset($c['view_mode']) ? $c['view_mode'] : 'teaser';
    $config['numitems'] = isset($c['numitems']) ? $c['numitems'] : '3';
    $sort = array(
      array(
        'method' => 'propertyOrderBy',
        'args' => array(
          'property' => 'title',
          'direction' => 'DESC',
        ),
      ),
    );
    $config['sort'] = isset($c['sort']) ? $c['sort'] : $sort;
    $config['pager'] = isset($c['pager']) ? $c['pager'] : '2';
    $config['pager_position'] = isset($c['pager_position']) ? $c['pager_position'] : '1';
    $config['pager_st_offset'] = isset($c['pager_st_offset']) ? $c['pager_st_offset'] : '175';
    $config['pager_st_time'] = isset($c['pager_st_time']) ? $c['pager_st_time'] : '900';
    $config['cache'] = isset($c['cache']) ? $c['cache'] : '1';
    $config['get_total'] = isset($c['get_total']) ? $c['get_total'] : '0';
    $config['status'] = isset($c['status']) ? $c['status'] : '1';
    $config['template'] = isset($c['template']) ? $c['template'] : NULL;
    $config['item_list'] = isset($c['item_list']) ? $c['item_list'] : '0';
    $config['tag'] = isset($c['tag']) ? $c['tag'] : NULL;
    $config['property_condition'] = isset($c['property_condition']) ? $c['property_condition'] : NULL;
    $config['field_condition'] = isset($c['field_condition']) ? $c['field_condition'] : NULL;

    $this->config = $config;

  }

  /**
   * Return the list as themed output.
   *
   * @param string $page
   *   The page we're on. For paged results, this will be '0' or an empty
   *   string for the first page of results, '1' for the second page, and so on.
   *
   * @return string
   *   The themed output.
   */
  public function getList($page = '') {

    $this->defaults();

    if ('0' === $this->config['numitems']) {
      return '';
    }

    // For bundle, we accept either an array or a json encoded string.
    if (!is_array($this->config['bundle'])) {
      $this->config['bundle'] = json_decode($this->config['bundle']);
      if (NULL === $this->config['bundle']) {
        $msg = 'Invalid json string in EntityLister::getItems().';
        watchdog('entity_lister', $msg);
        return '';
      }
    }

    $type = $this->config['type'];

    // Check if the user can view the items.
    $access = $this->accessCheck();
    if (!$access[0]) {
      return '';
    }
    elseif ('node' == $type) {
      // Bundles the user can view.
      $this->config['bundle'] = $access[1];
    }

    // Always require one or more bundles.
    if (empty($this->config['bundle'])) {
      return '';
    }

    if ('1' === $this->config['cache']) {
      $cache_config = $this->config;
      $cache_config['page'] = $page;
      // Determine the cache ID.
      $cid = $this->cid($cache_config);
      // Return the cached data if we have any.
      if ($cachedata = cache_get($cid)) {
        if ('2' === $this->config['pager']) {
          $this->pager();
        }
        return $cachedata->data;
      }
    }

    $this->bundleFields = array();
    // Build an array containing all fields in all the bundles.
    foreach ($this->config['bundle'] as $value) {
      $this->bundleFields += field_info_instances($type, $value);
    }

    /* Tell EntityFieldQuery and the entity's view function what page
     * we're on. We can set this any time before we execute the query.
     */
    $_GET['page'] = $page;

    $query = $this->query();
    $result = $query->execute();

    if (!$result || !isset($result[$type])) {
      return '';
    }
    else {
      $ids = array_keys($result[$type]);
      $entities = entity_load($type, $ids);
    }

    if ($entities) {
      $entities = $this->safeLoop($entities);
      $output = $this->theme($entities);
    }
    else {
      $output = '';
    }

    // In the cache table, the cid has a max of 255 chars, so in the unlikely
    // event our cid is longer, we can't cache the list.
    if ('1' === $this->config['cache']) {
      if (strlen($cid) > 255) {
        watchdog('entity_lister', 'A cid is greater than 255 characters in entity_liseter.class.inc.');
      }
      else {
        cache_set($cid, $output, 'cache', CACHE_TEMPORARY);
      }
    }

    return $output;

  }

  /**
   * For the standard pager, determine what page we're on.
   *
   * @return string
   *   The page number, e.g. '2'.
   */
  public function getPageNum() {

    return isset($_GET['page']) ? $_GET['page'] : '';

  }

  /**
   * Build the query.
   *
   * @return object
   *   The query.
   */
  protected function query() {

    extract($this->config);

    $query = new EntityFieldQuery();
    $query->entityCondition('entity_type', $type);

    if ($tag) {
      $query->addTag($tag);
    }

    // Don't display the host entity in its own list.
    if (!empty($this->hostItem)) {
      if ($type == $this->hostItem[3]) {
        if (!isset($bundle) || in_array($this->hostItem[2], $bundle)) {
          $query->entityCondition('entity_id', $this->hostItem[0], '!=');
        }
      }
    }

    $prop_info = entity_get_property_info($type);
    $props = $prop_info['properties'];

    // Deal with inconsistencies re: how entities store bundle data.
    // Checking the type property works seemingly for everything except
    // taxonomy_term bundles.
    if (isset($props['type'])) {
      $query->propertyCondition('type', $bundle);
    }
    // This entityCondition works in some cases, taxonomy_term being one
    // of those cases.
    elseif ('taxonomy_term' == $type) {
      $query->entityCondition('bundle', $bundle);
    }

    // Not all entity types have a status property.
    if (isset($props['status']) && 'all' !== $status) {
      $query->propertyCondition('status', $status, '=');
    }

    // For any property conditions besides type and status.
    if ($property_condition) {
      foreach ($property_condition as $key => $val) {
        $column = isset($val['column']) ? $val['column'] : NULL;
        $value = isset($val['value']) ? $val['value'] : NULL;
        $operator = isset($val['operator']) ? $val['operator'] : NULL;
        $query->propertyCondition($column, $value, $operator);
      }
    }

    if ($field_condition) {
      foreach ($field_condition as $key => $val) {
        $field = $val['field'];
        $column = isset($val['column']) ? $val['column'] : NULL;
        $value = isset($val['value']) ? $val['value'] : NULL;
        $operator = isset($val['operator']) ? $val['operator'] : NULL;
        $delta_group = isset($val['delta_group']) ? $val['delta_group'] : NULL;
        $language_group = isset($val['language_group']) ? $val['language_group'] : NULL;
        $query->fieldCondition($field, $column, $value, $operator, $delta_group, $language_group);
      }
    }

    foreach ($sort as $value) {
      $method = $value['method'];
      $args = $value['args'];
      $direction = isset($args['direction']) ? $args['direction'] : 'ASC';
      if ('propertyOrderBy' == $method) {
        $query->propertyOrderBy($args['property'], $direction);
      }
      elseif ('fieldOrderBy' == $method) {
        $query->fieldOrderBy($args['field'], $args['column'], $direction);
      }
    }

    if ($this->config['get_total']) {
      // Count the total number of items matching the conditions.
      $result = $query->execute();
      if ($result && isset($result[$type])) {
        $this->total = count($result[$type]);
      }
      else {
        $this->total = NULL;
      }
    }

    if ('0' !== $pager) {
      $query->pager($numitems, $this->delta);
    }
    else {
      $query->range(0, $numitems);
    }

    return $query;
  }

  /**
   * Determine a cache ID.
   *
   * @param array $cache_config
   *   Data used to determine the cache ID.
   *
   * @return string
   *   The cid.
   */
  protected function cid($cache_config) {

    extract($cache_config);

    // Use the list config and other data to build the cache ID.
    $cid['type'] = $type;
    // Besides the type, it doesn't matter in what order we add things
    // to $cid, as long as we keep it consistent.
    $cid['bundle'] = implode(',', $bundle);
    $cid['view_mode'] = $view_mode;
    $cid['numitems'] = $numitems;
    $cid['pager'] = $pager;
    $cid['pager_position'] = $pager_position;
    $cid['template'] = $template ? $template : '';
    $cid['item_list'] = $item_list ? '0' : '1';
    $cid['status'] = $status;
    // We also need the page for uniqueness.
    $cid['page'] = $page;
    $cid['tag'] = $tag ? $tag : '';
    /* We use MD5 hashes to limit some parts of the cid to 32 chars, with the
     * goal of keeping the entire cid to under 255 characters.  Note that we
     * could hash the entire cid except for the type if we wanted.
     */
    $cid['sort'] = hash('md5', json_encode($sort));
    if ($property_condition) {
      $property_condition = hash('md5', json_encode($property_condition));
    }
    $cid['property_condition'] = $property_condition ? $property_condition : '';
    if ($field_condition) {
      $field_condition = hash('md5', json_encode($field_condition));
    }
    $cid['field_condition'] = $field_condition ? $field_condition : '';

    // The cache must be unique for each set of roles a user might have.
    global $user;
    $roles = $user->roles;
    ksort($roles);
    $roles = array_keys($roles);
    $roles = implode(',', $roles);
    // The cache will be shared by users who have the same set of roles.
    $cid['roles'] = $roles;

    // Add the language.
    global $language;
    $cid['language'] = $language->language;

    $cid = 'entity_lister:' . implode(':', $cid);

    return $cid;
  }

  /**
   * Theme the list.
   *
   * @param array $entities
   *   The entities to list.
   *
   * @return string
   *   The themed output.
   */
  protected function theme($entities) {

    $pager = $this->config['pager'];
    $pager_position = $this->config['pager_position'];

    if ('2' === $pager) {
      $this->pager();
    }
    if ('0' !== $pager) {
      $vars['above'] = '0' === $pager_position || '2' == $pager_position;
      $vars['below'] = '1' === $pager_position || '2' == $pager_position;
    }
    else {
      $vars['above'] = $vars['below'] = FALSE;
    }

    $items = $this->renderItems($entities);

    if ('1' == $this->config['item_list']) {
      $vars['list'] = theme('item_list', array('items' => $items));
    }
    else {
      $vars['list'] = implode("", $items);
    }
    $vars['pager'] = theme('pager', array('element' => $this->delta));
    $vars['total'] = $this->config['get_total'] ? $this->total : NULL;
    // We might want access to the config vars in the template.
    $vars['config'] = $this->config;

    $output = theme('entity_lister_list', $vars);

    return $output;
  }

  /**
   * Render list items.
   *
   * @param array $entities
   *   The entities to list.
   *
   * @return array
   *   The rendered items.
   */
  protected function renderItems($entities) {

    $type = $this->config['type'];
    $viewfunc = $type . '_view';

    if (function_exists($viewfunc)) {
      foreach ($entities as $entity) {
        $items[] = drupal_render($viewfunc($entity, $this->config['view_mode']));
      }
    }
    elseif (function_exists('theme_entity_lister_' . $type)) {
      // How to provide a custom theme function in cases when an entity type
      // has no view function, as is the case with the taxonomy_vocabulary type.
      // See theme_entity_lister_taxonomy_vocabulary() for an example.
      $vars['entities'] = $entities;
      $items = theme('entity_lister_' . $type, $vars);
    }
    else {
      $msg = 'No view function or theme function in EntityLister::renderItems().';
      watchdog('entity_lister', $msg);
      // No view function and no theme function, so just list the titles.
      foreach ($entities as $entity) {
        $items[] = $entity->name;
      }
    }

    return $items;
  }

  /**
   * Activate the ajax pager.
   */
  protected function pager() {

    $path = drupal_get_path('module', 'entity_lister');
    drupal_add_js($path . '/ajax-pager.js');

    $js_settings[$this->delta] = array(
      'config' => $this->config,
      'delta' => $this->delta,
      'flag' => $this->flag,
      'host_item' => $this->hostItem,
      'list_class' => get_class($this),
    );
    drupal_add_js(array('entity_lister' => $js_settings), 'setting');

  }

  /**
   * Wrap the list in an element for rendering.
   *
   * @param string $list
   *   The themed list output.
   *
   * @return array
   *   The element.
   */
  public function element($list) {

    if (!empty($this->flag)) {
      $f = str_replace('_', '-', $this->flag) . '-';
    }
    else {
      $f = '';
    }

    // The AJAX pager uses this id.
    $id = 'entity-lister-' . check_plain($f . $this->delta);

    $element['entity-lister-' . $this->delta] = array(
      '#prefix' => '<div id="' . $id . '">',
      '#suffix' => '</div>',
      '#markup' => $list,
    );

    return $element;

  }

  /**
   * Check if the user can view the list items.
   *
   * @return array
   *   Info about the user's access to this type and its bundles.
   *   $info[0] bool -- Can we view the type?
   *   $info[1] array -- For nodes, the bundles we can view.
   */
  protected function accessCheck() {

    $info = array(TRUE, array());

    if ('node' == $this->config['type']) {

      if (!user_access('access content')) {
        $info[0] = FALSE;
      }
      else {
        // Add the node_view_permissions integration.
        $bundle = $this->nodeViewPermissions();
        $info[1] = $bundle;
      }

    }
    elseif ('user' == $this->config['type']) {
      if (!user_access('access user profiles')) {
        $info[0] = FALSE;
      };
    }
    elseif ('comment' == $this->config['type']) {
      if (!user_access('access comments')) {
        $info[0] = FALSE;
      };
    }
    return $info;
  }

  /**
   * Add the node_view_permissions integration.
   *
   * @return array
   *   Bundles the user is allowed to view.
   */
  protected function nodeViewPermissions() {

    $bundle = $this->config['bundle'];

    if (module_exists('node_view_permissions')) {
      $nvp_bundles = node_view_permissions_get_configured_types();
      foreach ($nvp_bundles as $key => $value) {
        $perm = 'view any ' . $value . ' content';
        if (!user_access($perm)) {
          $key = array_search($value, $bundle);
          unset($bundle[$key]);
        };
      }
    }
    return $bundle;

  }

  /**
   * Prevent infinite loops.
   *
   * Prevent display of Entity List fields in list items.
   *
   * @param array $entities
   *   The entities in the list.
   *
   * @return array
   *   Entities.
   */
  protected function safeLoop($entities) {

    $has = FALSE;

    foreach ($this->bundleFields as $value) {
      // Check if any of the bundles have an Entity List field.
      if ('entity_list_field' == $value['widget']['module']) {
        $fn = $value['field_name'];
        $has[$fn] = $fn;
      }
    }

    if ($has) {
      foreach ($has as $name) {
        // We need to remove the list fields from the items.
        foreach ($entities as $key => $entity) {
          if (isset($entity->{$name})) {
            unset($entity->{$name});
          }
          $entities[$key] = $entity;
        }
      }
    }

    return $entities;

  }

  /**
   * Helper function for adding a class to the allowed classes array.
   *
   * This is necessary for the AJAX pager to function, but is harmless in any
   * case.  If you are sure the AJAX pager will never be used on your site, you
   * don't have to do this.
   *
   * Modules extending EntityLister should add their class to the allowed
   * classes array, like so:
   *
   * EntityLister::addAllowedClass('MyClass');
   *
   * A good place to do this is in your hook_enable() call in your install file.
   *
   * @param string $class
   *   The class to allow.
   */
  public function addAllowedClass($class) {
    $allowed_classes = variable_get('entity_lister_allowed_classes', array());
    $allowed_classes[$class] = $class;
    variable_set('entity_lister_allowed_classes', $allowed_classes);
  }

  /**
   * Helper function for removing a class from the allowed classes array.
   *
   * Modules extending EntityLister should remove their class in hook_uninstall
   * like so:
   *
   * EntityLister::removeAllowedClass('MyClass');
   *
   * @param string $class
   *   The class to disallow.
   */
  public function removeAllowedClass($class) {
    $allowed_classes = variable_get('entity_lister_allowed_classes', array());
    if (isset($allowed_classes[$class])) {
      unset($allowed_classes[$class]);
    }
    variable_set('entity_lister_allowed_classes', $allowed_classes);
  }

}
